import java.util.prefs.Preferences;
import java.util.prefs.BackingStoreException;
import java.io.File;
import javax.swing.JOptionPane;

/**
 * Manages access to persistent application preferences.
 * 
 * @author David Foster <davidfstr@gmail.com>
 */
public class ImageFilterDemoPrefs {
	private final static boolean VERBOSE = false;
	
	private final static int SUPPORTED_PREFS_MAJOR_VERSION = 1;
	private final static int SUPPORTED_PREFS_MINOR_VERSION = 2;
	
	private final static String PREFS_DOMAIN = "ImageFilterDemo";
	private final static String ATTR_PATH_OF_LAST_SELECTED_INPUT_FILE = "lastSelectedInputFile.path";
	private final static String ATTR_PATH_OF_LAST_SELECTED_OUTPUT_DIR = "lastSelectedOutputDir.path";
	private final static String ATTR_NAME_OF_LAST_CHOSEN_OUTPUT_FORMAT = "lastChosenOutputFormat.name";
	
	private final static String NODE_PREFS_VERSION = "PrefsVersion";
	private final static String ATTR_PREFS_MINOR_VERSION = "minor";
	private final static String ATTR_PREFS_MAJOR_VERSION = "major";
	
	// ### INIT ###
	
	private static Preferences prefs = null;
	
	static {
		if (VERBOSE) {
			System.err.println("Loading preferences from persistent storage...");
		}
		
		// Load preferences from persistent storage
		boolean needToInitPrefs;
		try {
			Preferences userPrefsRoot = Preferences.userNodeForPackage(java.lang.Object.class);
			try {
				needToInitPrefs = userPrefsRoot.nodeExists(PREFS_DOMAIN);
			} catch (BackingStoreException e) {
				System.err.println("Error reading preferences from persistent storage:");
				e.printStackTrace(System.err);
				System.err.println("");
				System.err.println("Resetting preferences to default values.");
				needToInitPrefs = true;
			}
			prefs = userPrefsRoot.node(PREFS_DOMAIN);
		} catch (SecurityException e) {
			// TODO: increase robustness; make program assume default values for
			// preferences when the real values cannot be accessed
			JOptionPane.showMessageDialog(
				null,
				"Access to this program's stored preferences "+
					"was denied (due to insufficient permissions).",
				"Preference Access Error",
				JOptionPane.ERROR_MESSAGE,
				null);
			System.exit(1);
			
			//return;				// XXX: javac won't let me use this
			needToInitPrefs = true;	// XXX: required since I can't insert the preceding line
		}
		
		
		if (needToInitPrefs) {
			// Initialize preferences to its initial state if it didn't exist previously
			initPrefs();
		}
		
		// Ensure that we know how to read the preferences, based on its version
		checkPrefsVersion();
		
		// Add shutdown hook to ensure that the changes to the 
		// preferences are saved when the VM shuts down
		Runtime.getRuntime().addShutdownHook(new Thread("Shutdown: Save Preferences") {
			public void run() {
				try {
					if (VERBOSE) {
						System.err.println("Saving preferences to persistent storage...");
					}
					
					flushNow();
				} catch (Throwable e) {
					System.err.println("Error while flushing preferences to storage:");
					e.printStackTrace(System.err);
				}
			}
		});
	}
	
	private final static void initPrefs() {
		/* Setup initial state of preferences */
		setPrefsMajorVersion(SUPPORTED_PREFS_MAJOR_VERSION);
		setPrefsMinorVersion(SUPPORTED_PREFS_MINOR_VERSION);
		
		// Flush initial state of preferences to persistent storage
		flush();
	}
	
	private final static void checkPrefsVersion() {
		int prefsMajorVersion = getPrefsMajorVersion();
		
		if (prefsMajorVersion < SUPPORTED_PREFS_MAJOR_VERSION) {
			// TODO: increase robustness; offer to upgrade the user's preferences
			JOptionPane.showMessageDialog(
				null,
				"The version of this program's stored preferences "+
					"is older than this version of ImageFilterDemo "+
					"knows how to access. ",
				"Preference Access Error",
				JOptionPane.ERROR_MESSAGE,
				null);
			System.exit(1);
			return;
		}
		
		if (prefsMajorVersion > SUPPORTED_PREFS_MAJOR_VERSION) {
			// TODO: increase robustness; have program offer
			// to run with the default preferences and offer to either
			// override or preserve the stored preferences
			JOptionPane.showMessageDialog(
				null,
				"The version of this program's stored preferences "+
					"is newer than this version of ImageFilterDemo "+
					"knows how to access.",
				"Preference Access Error",
				JOptionPane.ERROR_MESSAGE,
				null);
			System.exit(1);
			return;
		}
	}
	
	private ImageFilterDemoPrefs() {}
	
	// ### PREFERENCES VERSION ###
	
	private static int getPrefsMajorVersion() {
		return prefs.node(NODE_PREFS_VERSION).getInt(ATTR_PREFS_MAJOR_VERSION, SUPPORTED_PREFS_MAJOR_VERSION);
	}
	
	private static int getPrefsMinorVersion() {
		return prefs.node(NODE_PREFS_VERSION).getInt(ATTR_PREFS_MINOR_VERSION, SUPPORTED_PREFS_MINOR_VERSION);
	}
	
	private static void setPrefsMajorVersion(int majorVersion) {
		prefs.node(NODE_PREFS_VERSION).putInt(ATTR_PREFS_MAJOR_VERSION, majorVersion);
	}
	
	private static void setPrefsMinorVersion(int minorVersion) {
		prefs.node(NODE_PREFS_VERSION).putInt(ATTR_PREFS_MINOR_VERSION, minorVersion);
	}
	
	// ### ACCESSORS ###
	
	/**
	 * @return		the input file that was last selected;
	 * 				or <tt>null</tt> if not defined
	 */
	public static File getLastSelectedInputFile() {
		String filepath = prefs.get(ATTR_PATH_OF_LAST_SELECTED_INPUT_FILE, null);
		return (filepath != null)
			? new File(filepath)
			: null;
	}
	
	public static void setLastSelectedInputFile(File lastSelectedInputFile) {
		String filepath = (lastSelectedInputFile != null)
			? lastSelectedInputFile.getAbsolutePath()
			: null;
		prefs.put(ATTR_PATH_OF_LAST_SELECTED_INPUT_FILE, filepath);
	}
	
	public static File getLastSelectedOutputDirectory() {
		String filepath = prefs.get(ATTR_PATH_OF_LAST_SELECTED_OUTPUT_DIR, null);
		return (filepath != null)
			? new File(filepath)
			: null;
	}
	
	public static void setLastSelectedOutputDirectory(File lastSelectedOutputDir) {
		String filepath = (lastSelectedOutputDir != null)
			? lastSelectedOutputDir.getAbsolutePath()
			: null;
		prefs.put(ATTR_PATH_OF_LAST_SELECTED_OUTPUT_DIR, filepath);
	}
	
	public static String getLastChosenOutputFormatName() {
		return prefs.get(ATTR_NAME_OF_LAST_CHOSEN_OUTPUT_FORMAT, null);
	}
	
	public static void setLastChosenOutputFormatName(String lastChosenOutputFormatName) {
		prefs.put(ATTR_NAME_OF_LAST_CHOSEN_OUTPUT_FORMAT, lastChosenOutputFormatName);
	}
	
	// ### MISC ###
	
	/**
	 * Attempts to flush any preference changes to persistent storage.
	 * 
	 * @return		whether the flush was successful
	 */
	public static boolean flush() {
		try {
			flushNow();
			return true;
		} catch (BackingStoreException e) {
			System.err.println("Error while incrementally flushing preferences to persistent store:");
			e.printStackTrace(System.err);
			return false;
		}
	}
	
	/**
	 * Forces any preference changes to be flushed to persistent storage immediately.
	 */
	private static void flushNow() throws BackingStoreException {
		if (prefs != null) {
			prefs.flush();
		}
	}
}